#!/usr/bin/env gimp-script-fu-interpreter-3.0
;; -----------------------------------------------------------------------------
;; script-fu-glow-layer:
;;   Generate a multi‑layered glow by cloning the source and applying
;;   successive blurs & spread (noise) with varying strength.
;;
;;   Find in Filters > Light and Shadow > Gaussian Glow
;;
;; Parameters:
;;   iterations         — The number of layers to create for constructing the glow.
;;                         • Lower values → faster, coarser; higher → slower, smoother
;;   initial-opacity    — Opacity (%) for the very first glow layer, which uses the
;;                         largest blur radius (most‑blurred).
;;                         • Lower values → subtler outer glow; higher → stronger.
;;   final-opacity      — Opacity (%) for the last layer
;;                         • Lower values → subtler inner glow; higher → stronger.
;;   opacity-exponent   — How the opacity values rise, the curve of the gradient
;;                         • Lower values → slow out fast in; higher → fast out slow in
;;   final-gauss        — Gaussian blur radius (pixels) for the last layer.
;;                         • Larger → broader halo; smaller → tighter glow.
;;   gauss-exponent     — How the blur values fall, the curve of the gradient
;;   final-spread       — Noise spread (pixels) on the last layer.
;;                         • Adds grain/texture; higher → more “fuzzy” glow.
;;   spread-exponent    — How the spread values fall, the curve of the gradient
;;   post-gauss         — Gaussian blur radius (pixels) for the composited layer.
;;                         • Used to soften pixelation caused by noise spread.
;;   keep-construction  — Boolean: TRUE keeps the temporary glow image
;;                         (and all intermediate layers).
;;   show-info          — Boolean: Outputs the dynamic values to the error console
;;                         (and displays the construction image as it is created).
;; -----------------------------------------------------------------------------

;; Copyright 2024, Mark Sweeney, Under GNU GENERAL PUBLIC LICENSE Version 3

;; Global variables
(define debug #f)

;; =============================================================================
;; Main
(define (script-fu-glow-layer image drawables
                              pass-count
                              initial-opacity
                              final-opacity
                              opacity-exponent
                              final-gauss
                              gauss-exponent
                              final-spread
                              spread-exponent
                              post-gauss
                              keep-construction
                              show-info)
  ;; Populate the global `config` alist with these GUI parameters for easy passing
  (set! config
    `((image . ,image)
      (drawables . ,drawables)
      (pass-count . ,pass-count)
      (initial-opacity . ,initial-opacity)
      (final-opacity . ,final-opacity)
      (opacity-exponent . ,opacity-exponent)
      (final-gauss . ,final-gauss)
      (gauss-exponent . ,gauss-exponent)
      (final-spread . ,final-spread)
      (spread-exponent . ,spread-exponent)
      (post-gauss . ,post-gauss)
      (keep-construction . ,keep-construction)
      (show-info . ,show-info)))
  (convert-gui-settings)
  ;; Call the main function
  (glow-layer))

;; -----------------------------------------------------------------------------

(script-fu-register-filter
  "script-fu-glow-layer"
  "Gaussian Glow..."
  "Generate a glow by applying successive blurs & scatter with varying strength"
  "Mark Sweeney"
  "Under GNU GENERAL PUBLIC LICENSE Version 3"
  "2025"
  "*"
  SF-ONE-DRAWABLE
  SF-ADJUSTMENT _"Nu_mber of Passes"  '(16 3 64 1 10 0 SF-SPINNER)
  SF-ADJUSTMENT _"Initial _Opacity"   '(5 1 100 1 10 0 SF-SPINNER)
  SF-ADJUSTMENT _"Final Opacit_y"     '(100 1 100 1 10 0 SF-SPINNER)
  SF-ADJUSTMENT _"O_pacity Exponent"  '(0.4 0.01 16 0.05 0.1 1 SF-SPINNER)
  SF-ADJUSTMENT _"Final G_auss"       '(1 0.1 499 1 10 1 SF-SPINNER)
  SF-ADJUSTMENT _"_Gauss Exponent"    '(2.2 0.01 16 0.01 0.05 1 SF-SPINNER)
  SF-ADJUSTMENT _"_Final Spread"      '(2 1 100 1 10 0 SF-SPINNER)
  SF-ADJUSTMENT _"Spread E_xponent"   '(2.2 0.01 16 0.01 0.05 1 SF-SPINNER)
  SF-ADJUSTMENT _"Pos_t Gauss"        '(0.75 0.0 16 0.01 0.05 1 SF-SPINNER)
  SF-TOGGLE     _"Keep Construct_ion"  0
  SF-TOGGLE     _"S_how Info"          0)

(script-fu-menu-register
  "script-fu-glow-layer"
  "<Image>/Filters/Light and Shadow/")

;; =============================================================================
;; Local Functions

;; Purpose: The top level procedure for the plug-in after the GUI phase
(define (glow-layer)
  (call 'glow-layer)
  (let* ((image (get-image))
         (active-layer (prepare-image-and-get-active-layer))
         (glow-layer-name (concat "glow-" (item-get-name active-layer)))
         (glow-layer (clone-layer image active-layer glow-layer-name))
         (glow-resize-and-copy (resize-layer-to-image-and-copy glow-layer))
         (tmp-image (edit-paste-as-new-image))
         (tmp-display (create-display? tmp-image))
         (tmp-layer (get-active-layers-vector tmp-image #t))
         (crop-glow-margin 64)
         (post-gauss (get 'post-gauss)))
      (create-glow-layers tmp-image tmp-layer)
      (transfer-image-to-layer tmp-image image glow-layer LAYER-MODE-ADDITION)
      (when (is-true? post-gauss) (post-process glow-layer post-gauss))
      (layers-alpha-crop image (list glow-layer) crop-glow-margin)
      (end-plug-in tmp-image tmp-display)))

;; Purpose: Format the raw Script-Fu GUI values for boolean readability
(define (convert-gui-settings)
  (call 'convert-gui-settings)
  (set-config 'keep-construction (is-true? (get 'keep-construction)))
  (set-config 'show-info (is-true? (get 'show-info))))

;; Purpose: Returns a new display if debugging or keeping the construction image
(define (create-display? tmp-image) (call 'create-display?)
  (if (or debug (get 'keep-construction))
    (display-new tmp-image)
    #f))

;; Purpose: Returns the active layer after validation and sets up the image
;;          ready to start glowing
(define (prepare-image-and-get-active-layer)
  (call 'prepare-image-and-get-active-layer)
  (let ((layer (ensure-layer (get 'drawables))))
    (when (item-is-group? layer)
      (exit-message "Please select a layer."))
    (gimp-image-undo-group-start (get-image))
    (gimp-selection-none (get-image))
    layer))

;; Purpose: Clean up the display or image created and end the plug-in progress
(define (end-plug-in tmp-image tmp-display)
  (call 'end-plug-in)
  (unless (get 'keep-construction)
    (if debug
      (gimp-display-delete tmp-display)
      (gimp-image-delete tmp-image)))
  (gimp-progress-end)
  (gimp-image-undo-group-end (get-image)))

;; Purpose: Calculate intial blur and noise size and create N layers
(define (create-glow-layers tmp-image tmp-layer)
  (call 'create-glow-layers)
  (let* ((pass-count (get 'pass-count))
         (info-str "     | Opacity | Gauss | Spread \n")
         (initial-gauss (calculate-initial-gauss-radius tmp-image)))
    (do ((i 0 (+ i 1)))
        ((>= i pass-count))
      (let* ((glow-params (calculate-glow-params i pass-count initial-gauss))
             (opacity (first-item glow-params))
             (gauss (second-item glow-params))
             (spread (third-item glow-params)))
        (gimp-progress-pulse)
        (gimp-progress-set-text (concat "Pass " (+ i 1) " of " pass-count))
        (set! info-str (format-info info-str (+ i 1) opacity gauss spread))
        (process-layer tmp-image tmp-layer opacity gauss spread)))
    (when (get 'show-info) (message info-str))
    (gimp-item-set-visible tmp-layer FALSE)))

;;Purpose: Returns the initial gaussian blur radius based on the image size
(define (calculate-initial-gauss-radius tmp-image)
  (call 'calculate-initial-gauss-radius)
  (let* ((geometric-mean (image-get-geometric-mean tmp-image))
         (initial-radius (* 0.25 geometric-mean)) ;; a quarter of the image size
         (radius-limit 500))
    (min initial-radius radius-limit)))

;;Purpose: Returns the calculated opacity, blur and spread parameters to apply
;;         to the current layer. Keeping some invariants inside loop for
;;         readability, returns a list (opacity gauss spread)
(define (calculate-glow-params i pass-count initial-gauss)
  (call 'calculate-glow-params)
  (let* ((initial-spread  (* initial-gauss 0.5))
         (initial-opacity (get 'initial-opacity))
         (final-opacity   (get 'final-opacity))
         (final-gauss     (get 'final-gauss))
         (final-spread    (get 'final-spread))
         (opacity-exp     (get 'opacity-exponent))
         (gauss-exp       (get 'gauss-exponent))
         (spread-exp      (get 'spread-exponent))
         (gauss-range   (- initial-gauss final-gauss))
         (spread-range  (- initial-spread final-spread))
         (opacity-range (- final-opacity initial-opacity))
         (norm-opac   (normalized-loop-progress opacity-exp i pass-count #t))
         (norm-gauss  (normalized-loop-progress gauss-exp i pass-count #f))
         (norm-spread (normalized-loop-progress spread-exp i pass-count #f))
         (opacity     (+ initial-opacity (* opacity-range norm-opac)))
         (gauss       (+ final-gauss     (* gauss-range norm-gauss)))
         (spread      (+ final-spread    (* spread-range norm-spread))))
    (list opacity gauss spread)))

;;Purpose: Duplicate the layer with opacity, apply GEGL blur of radius gauss,
;;         then spread pixels by GEGL noise-spread (clamped to min-spread).
(define (process-layer tmp-image tmp-layer opacity gauss spread)
  (call 'process-layer)
  (let ((loop-layer (duplicate-layer tmp-image
                                     tmp-layer
                                     (concat "O:" (float->int opacity) "|"
                                             "G:" (float->int gauss) "|"
                                             "S:" (float->int spread))
                                     opacity
                                     LAYER-MODE-NORMAL
                                     TRUE)))
    (gegl-gaussian-blur loop-layer
                        LAYER-MODE-REPLACE
                        opaque
                        gauss gauss
                        "auto"
                        #t #f)
    (gegl-noise-spread loop-layer
                       LAYER-MODE-REPLACE
                       opaque
                       spread spread
                       (random-seed)
                       #t #f)))

;;Purpose: Apply a blur after merging the glow layers to any soften pixellation
(define (post-process layer gauss)
  (call 'post-process)
  (gegl-gaussian-blur layer
                      LAYER-MODE-REPLACE
                      opaque
                      gauss gauss
                      "auto"
                      #t #f))

;;Purpose: Formats a string for clear output as a table in the Error Console
(define (format-info info-str pass opacity gauss-blur spread)
  (call 'format-info )
  ;;  Each rule is a list of padding pairs for left and right alignment.
  ;;  The format is: ((<10) (<100) (>=100)), where each pair is (left right).
  ;;  Used by `format-cell` to align values for table-style output.
  (let ((pass-rules   '((0 3) (0 1) (0 0)))
        (opac-rules   '((4 5) (2 5) (0 5)))
        (gauss-rules  '((4 2) (2 2) (0 2)))
        (spread-rules '((4 0) (2 0) (0 0))))
    (concat
      info-str
      (format-cell pass 0 pass-rules "| ")
      (format-cell opacity 1 opac-rules "| ")
      (format-cell gauss-blur 1 gauss-rules "| ")
      (format-cell spread 1 spread-rules "")
      "\n")))

;; =============================================================================
;; Library Functions

;; Purpose: Utility operations and helper functions for Script-Fu plugins
;;
;; This is a collection of utility functions for common operations
;; in Script-Fu plugins, including path management, list manipulation, messaging,
;; and control flow utilities. These functions can be loaded into any Script-Fu
;; plugin to provide reusable utilities, helping to streamline the development of plugins.
;; =============================================================================

(define opaque 100)

(define config '())

(define warning #t)

;; Purpose: Ensures the given input is a list.
;; Converts a vector to a list or wraps other single values in a list if needed.
(define (ensure-list input)
  (cond
    ;; If the input is already a list, return it as-is
    ((list? input)
     input)
    ;; If the input is a vector, convert it to a list
    ((vector? input)
     (vector->list input))
    ;; If the input is null, return an empty list
    ((null? input)
     '())
    ;; Otherwise, wrap the input in a list
    (else
     (list input))))

;; Purpose: Ensures the given input is a vector.
;; Converts a list to a vector or wraps other single values in a vector if needed.
(define (ensure-vector input)
  (cond
    ;; If the input is already a vector, return it as-is
    ((vector? input)
     input)
    ;; If the input is a list, convert it to a vector
    ((list? input)
     (list->vector input))
    ;; If the input is null, return an empty vector
    ((null? input)
     (vector))
    ;; Otherwise, wrap the input in a vector
    (else
     (vector input))))

;; Purpose: Returns the first item of a list or vector. Warns if the input is
;;          invalid or empty.
(define (first-item collection)
  (cond
    ;; Handle non-empty lists
    ((and (list? collection) (not (null? collection)))
     (list-ref collection 0))
    ;; Handle non-empty vectors
    ((and (vector? collection) (> (vector-length collection) 0))
     (vector-ref collection 0))
    ;; Invalid or empty input
    (else
     (begin
       (warning-message "first-item: Expected a non-empty list or vector, but received: " collection)
       #f))))

;; Purpose: Abstract the 'cadr' function into a descriptive term.
;; Returns the second item of a list or vector. Warns if the input is invalid or too short.
(define (second-item collection)
  (cond
    ;; Handle non-empty lists with at least two elements
    ((and (list? collection) (> (length collection) 1))
     (list-ref collection 1))
    ;; Handle non-empty vectors with at least two elements
    ((and (vector? collection) (> (vector-length collection) 1))
     (vector-ref collection 1))
    ;; Warn if the collection is too short
    ((or (null? collection)
         (and (list? collection) (<= (length collection) 1))
         (and (vector? collection) (<= (vector-length collection) 1)))
     (begin
       (warning-message "second-item: Collection has fewer than 2 elements.")
       #f))
    ;; Invalid input
    (else
     (begin
       (warning-message "second-item: Expected a list or vector, but received: " collection)
       #f))))

;; Purpose: Abstract the 'caddr' function into a descriptive term.
;; Returns the third item of a list or vector. Warns if the input is invalid or too short.
(define (third-item collection)
  (cond
    ;; Handle non-empty lists with at least three elements
    ((and (list? collection) (> (length collection) 2))
     (list-ref collection 2))
    ;; Handle non-empty vectors with at least three elements
    ((and (vector? collection) (> (vector-length collection) 2))
     (vector-ref collection 2))
    ;; Warn if the collection is too short
    ((or (null? collection)
         (and (list? collection) (<= (length collection) 2))
         (and (vector? collection) (<= (vector-length collection) 2)))
     (begin
       (warning-message "third-item: Collection has fewer than 3 elements.")
       #f))
    ;; Invalid input
    (else
     (begin
       (warning-message "third-item: Expected a list or vector, but received: " collection)
       #f))))

;; Purpose: Abstract the 'cdr' function into a descriptive term, remove the head.
;; Warns if the input is not a list or is empty.
(define (list-remove-head list)
  (if (and (list? list) (not (null? list)))
    (cdr list)
    (begin
      (warning-message "list-remove-head: Expected a non-empty list, but received: " list)
      #f)))

;; Purpose: Unwrap a single-element () list or handle empty lists. Warns if the input is not a valid list or not a single-element list.
(define (list->item list)
  (cond
    ((null? list)
     (debug-message "list->item: Input is an empty list, returning '()")
     '())
    ((and (list? list) (not (list? (car list))) (= 1 (length list)))
     (car list))
    (else
     (debug-message "list->item: Input is not a single-element list, returning the list.")
     list)))

;; Purpose: Display a message in GIMP's message console.
(define (message . items)
  (gimp-message (apply concat items)))

;; Purpose: Display a message in a dialog box.
(define (message-box msg)
  (gimp-message-set-handler 0)
  (message msg "\n")
  (gimp-message-set-handler 2))

(define (call . items)
  (when debug (message "* (" (apply concat items) ")")))

;; Purpose: Display a debug message.
(define (debug-message . items)
  (when debug (message "> " (apply concat items))))

;; Purpose: Display a warning message.
(define (warning-message . items)
  (if warning
    (message "Warning: " (apply concat items)))
    #f)

;; Purpose: Serialize a list of items into a single string, handling numbers,
;;          booleans, and nested lists.
(define (list->string list)
  (if (list? list)
      (string-append "list: \n" (string-join (map serialize-item list) "\n"))
      (warning-message "list->string: Input is not a list!")))

;; Purpose: Converts various Scheme data types (lists, vectors, pairs, etc.)
;;          into a string representation
(define (serialize-item item)
  (cond
    ((and (list? item) (null? item)) "\"\"")           ; Empty list
    ((and (string? item) (string=? item "")) "\"\"")   ; Empty string
    ((list? item) (list->string item))                ; Nested list
    ((vector? item)                                   ; Handle vectors
     (string-append "#("
                    (string-join (map serialize-item (vector->list item)) " ")
                    ")"))
    ((pair? item)                                     ; Handle pairs
     (string-append "("
                    (serialize-item (car item))
                    " . "
                    (serialize-item (cdr item))
                    ")"))
    ((number? item) (number->string item))            ; Numbers
    ((symbol? item) (symbol->string item))            ; Symbols
    ((boolean? item) (if item "#t" "#f"))             ; Booleans
    ((string? item) item)                             ; Strings
    (else (warning-message "serialize-item: Unsupported item type!" item))))

;; Purpose: Concatenate multiple items into a single string, handling numbers,
;;          booleans, lists, and other types.
(define (concat . items)
  (apply string-append (map serialize-item items)))

;; Purpose: Join a list of strings with a delimiter
(define (string-join list delimiter)
  (let loop ((remaining list)
             (result ""))
    (cond
      ((null? remaining) result)  ;; Base case: no more elements
      ((null? (cdr remaining))    ;; Last element: add without delimiter
       (string-append result (car remaining)))
      (else
       (loop (cdr remaining)
             (string-append result (car remaining) delimiter))))))

;; Purpose: Returns the value from a global config alist pair if the parameter exists,
;;          otherwise returns a special "not found" marker.
(define (get parameter)
  (let ((result (assoc parameter config)))
    (if result
      (begin
        (when debug (message "> " parameter " : " (cdr result)))
        (cdr result))
        'not-found))) ;; Use a distinct value to indicate "not found"

;; Purpose: Returns the value from a global config alist pair if the parameter exists,
;;          otherwise returns a special "image-not-found" marker.
(define (get-image)
  (let ((result (assoc 'image config)))
    (if result
      (cdr result)
      'image-not-found)))

;; Purpose: Sets a key-value pair in the global `config` alist.
;; If the key already exists, updates its value; otherwise, adds the key-value pair.
(define (set-config key value)
  (let ((existing-pair (assoc key config)))
    (if existing-pair
        ;; Update the existing key's value
        (set-cdr! existing-pair value)
        ;; Add a new key-value pair to the config
        (set! config (cons (cons key value) config)))))

;; Purpose: Display a message in a box and exit the script.
(define (exit-message msg)
  (message-box msg)
  (quit))

;; Purpose: Check if an item is valid, returning #t or #f.
(define (item-is-valid? item)
  (if (number? item)
    (= (list->item (gimp-item-id-is-valid item)) 1)
    #f))

;; Purpose: Return the name of an item.
(define (item-get-name item)
  (if (not (= item 0))
    (if (item-is-valid? item)
      (list->item (gimp-item-get-name item))
      "invalid item")
    "root"))

;; Purpose: Returns the parent of an item if valid, otherwise returns #f
(define (item-get-parent item)
  (if (item-is-valid? item)
    (list->item (gimp-item-get-parent item))
    0))

;; Purpose: Returns #t if the item is a group, otherwise returns #f
(define (item-is-group? item)
  (and (item-is-valid? item)
       (= 1 (list->item (gimp-item-is-group item)))))

;; Purpose: Returns the items tree position or #f if the item is invalid
(define (get-item-tree-position image item)
  (if (item-is-valid? item)
    (list->item (gimp-image-get-item-position image item))
    #f))

;; Purpose: Returns #t if the item is a layer, otherwise returns #f
(define (item-is-layer? item)
  (and (item-is-valid? item)
       (= 1 (list->item (gimp-item-id-is-layer item)))))

;; Purpose: Returns #t if the item is a mask, otherwise returns #f
(define (item-is-layer-mask? item)
  (and (item-is-valid? item)
       (= 1 (list->item (gimp-item-id-is-layer-mask item)))))

;; Purpose: Returns #t if there is a selection
(define (has-a-selection? image)
  (= 0 (list->item (gimp-selection-is-empty image))))

;; Purpose: Returns #t for non-zero values, otherwise #f
(define (is-true? value)
  (not (= value 0)))

;; Purpose: Validate the range of a parameter
(define (validate-range value min max)
  (if (or (< value min) (> value max))
      (warning-message "Value out of range")
      value))

;; Purpose: Converts a bbox vector to a string in the format "top-left (x1,y1) bottom-right (x2,y2)"
(define (bbox->string bbox)
  (let* ((bbox-list (vector->list bbox))
         (x1 (list-ref bbox-list 0))
         (y1 (list-ref bbox-list 1))
         (x2 (list-ref bbox-list 2))
         (y2 (list-ref bbox-list 3)))
    (string-append "top-left (" (number->string x1) "," (number->string y1) ") "
                   "bottom-right (" (number->string x2) "," (number->string y2) ")")))

;; Purpose: Returns a copy of a given vector
(define (vector-copy vec)
  ;; Create a new vector with the same length as the original
  (let ((new-vec (make-vector (vector-length vec))))
    ;; Copy each element from the original vector to the new one
    (do ((i 0 (+ i 1)))
        ((= i (vector-length vec)) new-vec) ; Return the new vector when done
      (vector-set! new-vec i (vector-ref vec i)))
    new-vec))

;; Purpose: Returns a random int for seeding a filter
(define (random-seed)
 (msrg-rand))

;; Purpose: Convert a (possibly inexact) real to an exact integer by
;;          truncating toward zero and dropping any “.0” inexactness.
(define (float->int f)
  (inexact->exact (truncate f)))

;; Purpose: Truncate a float to N decimal places
(define (truncate-float f N)
  (if (is-true? N)
    (let ((multiplier (expt 10 N)))
      (/ (truncate (* f multiplier)) multiplier))
    (float->int f)))

;; Purpose: Returns a factor in [0, 1] based on the index, number of steps,
;;          and the exponent. Maps a loop index to a normalized non-linear
;;          progress value. An exponent of:
;;            1   - linear progression
;;            >1  - slow start, fast end
;;            <1  - fast start, slow end
;;          If 'invert' is true, returns 1 to 0 instead of 0 to 1.
(define (normalized-loop-progress exponent index steps invert)
  (let* ((denominator (- steps 1))            ;; intervals = steps − 1
         (u           (/ index denominator))) ;; normalized 0…1
    (if invert
      (- 1 (expt (- 1 u) exponent))
      (expt (- 1 u) exponent))))

;; Purpose: Returns a string padded value
(define (pad-str val left-pad right-pad)
  (let* ((str   (number->string val))
         (left  (make-string left-pad  #\space))
         (right (make-string right-pad #\space)))
    (string-append left str right)))

;; Purpose: Returns a string representing the padded, truncated, and suffixed
;;          value, suitable for fixed-width column alignment (e.g., table output)
;;
;; Parameters:
;;   value        - the numeric value to format
;;   truncate-to  - number of decimal places
;;   rules        - a list of (left right) padding pairs, selected
;;                  based on whether the value is <10, <100, or >=100
;;   suffix       - a string appended after the padded value, e.g "|"
(define (format-cell value truncate-to rules suffix)
  (let* ((val (truncate-float value truncate-to))
         (pads (cond ((< val 10)  (first-item rules 0))
                     ((< val 100) (second-item rules 1))
                     (else        (third-item rules 2))))
         (left  (first-item pads))
         (right (second-item pads)))
    (string-append (pad-str val left right) suffix)))

;; =============================================================================
;; Purpose: Layer Utilities for Script-Fu Plugins
;;
;; This module provides a collection of layer related functions
;; =============================================================================

;; Purpose: Returns a layer if given a mask of a layer, a vector of layers,
;;          a list of layers or a single layer ID.
;;          Typical use is to filter the user selection of drawables to be a
;;          single layer and not a mask.
(define (ensure-layer item)
  (cond
    ;; Case 1: `item` is a vector or a list
    ((or (vector? item) (list? item))
     (let ((layer (first-item item)))
       (if (item-is-layer-mask? layer)
           (layer-from-mask layer)
           layer)))
    ;; Case 2: `item` is a single layer mask
    ((item-is-layer-mask? item)
     (layer-from-mask item))
    ;; Case 3: `item` is a single layer
    ((item-is-layer? item)
     item)
    ;; Default case: Return #f if it does not match the above.
    (else #f)))

;; Purpose: Returns a layers mask, otherwise returns #f
(define (layer-from-mask item)
  (if (item-is-valid? item)
    (let ((mask (list->item (gimp-layer-from-mask item))))
      (if (= mask -1)
        #f
        mask))
    #f))

;; Purpose: Returns a list of the active layers or the first item if `first` is #t
;;          Returns #f if no layers are selected.
(define (get-active-layers-list image first)
  (let ((layers (list->item (gimp-image-get-selected-layers image))))
    (if first
      (if (> (vector-length layers) 0) ;; Check if the vector is non-empty
        (vector-ref layers 0)        ;; Return the first layer
        #f)                          ;; Return #f if no layers
      (if (> (vector-length layers) 0) ;; Check if the vector is non-empty
        (vector->list layers)        ;; Convert to a list and return
        #f))))                       ;; Return #f if no layers

;; Purpose: Returns a vector of the active layers or the first item if `first` is #t
;;          Returns #f if no layers are selected.
(define (get-active-layers-vector image first)
  (let ((layers (list->item (gimp-image-get-selected-layers image))))
    (if first
      (if (> (vector-length layers) 0) ;; Check if the vector is non-empty
        (vector-ref layers 0)        ;; Return the first layer
        #f)                          ;; Return #f if no layers
      (if (> (vector-length layers) 0) ;; Check if the vector is non-empty
        layers
        #f))))                       ;; Return #f if no layers

;; Purpose: Returns a new layer from a drawable
(define (layer-new-from-drawable layer image)
  (list->item (gimp-layer-new-from-drawable layer image)))

;; Purpose: Returns a duplicate of a layer and assigns specific attributes
(define (duplicate-layer image src-drawable name opacity mode visible)
  (let ((layer (layer-new-from-drawable src-drawable image)))
    (gimp-image-insert-layer image
                             layer
                             (item-get-parent src-drawable)
                             (get-item-tree-position image src-drawable))
    (gimp-layer-set-opacity layer opacity)
    (gimp-layer-set-mode layer mode)
    (gimp-item-set-name layer name)
    (gimp-item-set-visible layer visible)
    layer))

;; Purpose: Returns a named clone of a layer
(define (clone-layer image src-drawable name)
  (let ((layer (layer-new-from-drawable src-drawable image)))
    (gimp-image-insert-layer image
                             layer
                             (item-get-parent src-drawable)
                             (get-item-tree-position image src-drawable))
    (gimp-item-set-name layer name)
    layer))

;; Purpose: Copies drawables to the buffer
(define (edit-copy src-drawable)
  (debug-message "edit-copy " (item-get-name src-drawable))
  (gimp-edit-copy (ensure-vector src-drawable)))

;; Purpose: Returns the first vector of the pasted buffer list
(define (edit-paste dst-drawable)
  (debug-message "edit-paste " (item-get-name dst-drawable))
  (first-item (gimp-edit-paste dst-drawable TRUE)))

;; Purpose: Returns a layer pasted from the copy buffer to a destination drawable,
;;          options to set opacity mode and visibility.
(define (paste-copied-to-drawable dst-drawable dst-image opacity mode visible)
  (let ((layer (first-item (edit-paste dst-drawable))))
    (gimp-floating-sel-anchor layer)
    (set! layer (ensure-layer (get-active-layers-list dst-image #t)))
    (gimp-layer-set-opacity layer opacity)
    (gimp-layer-set-mode layer mode)
    (gimp-item-set-visible layer visible)
    layer))

;; Purpose: Get the x offset of the drawable from the origin
(define (drawable-get-offset-x layer)
  (first-item (gimp-drawable-get-offsets layer)))

;; Purpose: Get the y offset of the drawable from the origin
(define (drawable-get-offset-y layer)
  (second-item (gimp-drawable-get-offsets layer)))

;; Purpose: Returns #t (true) if the layer has an alpha
(define (layer-has-alpha? layer)
  (if (> (list->item (gimp-drawable-has-alpha layer)) 0)
    (begin
      (debug-message "layer has alpha : " (item-get-name layer))
      #t)
    (begin
      (debug-message "layer has no alpha : " (item-get-name layer))
      #f)))

;; Purpose: Crops a list of layers to the alpha area with a safe border distance
(define (layers-alpha-crop image layers border)
  (gimp-context-push)
  ;; Process each layer
  (for-each
    (lambda (layer)
      (let* ((old-off-x (drawable-get-offset-x layer))
             (old-off-y (drawable-get-offset-y layer))
             (bbox (if (layer-has-alpha? layer)
                     (begin
                       (select-layer-outline image layer #f)
                       (get-expanded-and-cropped-bbox image border))
                     #f)))
        ;; Resize the layer
        (when bbox
          (gimp-layer-resize layer
                             (bbox-width bbox)
                             (bbox-height bbox)
                             (- old-off-x (bbox-off-x bbox))
                             (- old-off-y (bbox-off-y bbox))))))
    (ensure-list layers))
    (gimp-selection-none image)
    (gimp-context-pop))

;; Purpose: Returns #t (true) if the layer has no alpha value to select, 0,0,0,0
;;          NOTE: gimp-image-select-item is slow on larger layers
(define (layer-has-selectable-alpha? image layer)
  (if (layer-has-alpha? layer)
    (begin
      (gimp-image-select-item image CHANNEL-OP-REPLACE layer)
      (has-a-selection? image))
    #f))

;; Purpose: Selects the alpha area of a layer, or inverted alpha,
(define (select-layer-outline image layer invert)
  (if (layer-has-selectable-alpha? image layer)
    (if invert
      (begin
        (gimp-image-select-item image CHANNEL-OP-REPLACE layer)
        (gimp-selection-invert image))
      (gimp-image-select-item image CHANNEL-OP-REPLACE layer))
    #f))

;; Purpose: Resize the layer to match the image size and copy to the buffer
(define (resize-layer-to-image-and-copy layer)
  (gimp-layer-resize-to-image-size layer)
  (edit-copy layer))
;; =============================================================================
;; Purpose: Image Handling Utilities
;;
;; This module provides a collection of image-related utility functions. These
;; functions can be used to retrieve information about images, such as file paths,
;; short file names, base names, and image states (e.g., whether the image is valid
;; or has been modified). The utilities are designed to help manage image-related
;; operations efficiently across various use cases.
;; =============================================================================

;; Purpose: Returns an image pasted from the clipboard or #f if no image is created
(define (edit-paste-as-new-image)
  (let ((image (list->item (gimp-edit-paste-as-new-image))))
    (if (> image 0)
      image
      #f)))

;; Purpose: Returns the image width
(define (image-get-width image)
  (list->item (gimp-image-get-width image)))

;; Purpose: Returns the image height
(define (image-get-height image)
  (list->item (gimp-image-get-height image)))

;; Purpose: Returns a new display
(define (display-new image)
  (list->item (gimp-display-new image)))

;; Purpose: Get the bounding box of the active image selection or return #f (false)
;;          if the selection is empty, the bounding box is a vector of four points.
;;          upper left, upper right, lower left, lower right.
(define (image-selection-get-bbox image)
  (let ((bounds (gimp-selection-bounds image)))
    (if (= (first-item bounds) 0)
      (begin
        (warning-message "image has no active selection, no bbox!")
        #f)
      (begin
        (let ((bbox (list->vector (list-remove-head bounds))))
          (debug-message "bbox of active selection: " (bbox->string bbox))
          bbox)))))

;; Purpose: Returns an expanded bounding box by a specified border amount.
;;          Handles both vector and list inputs, always returning a vector.
(define (expand-bbox bbox-in border)
  ;; Create a copy of the bbox vector and expand it
  (let* ((bbox (ensure-vector bbox-in))
         (expanded-bbox (vector-copy bbox)))
    (vector-set! expanded-bbox 0 (- (vector-ref expanded-bbox 0) border)) ; upper left
    (vector-set! expanded-bbox 1 (- (vector-ref expanded-bbox 1) border)) ; upper right
    (vector-set! expanded-bbox 2 (+ (vector-ref expanded-bbox 2) border)) ; lower left
    (vector-set! expanded-bbox 3 (+ (vector-ref expanded-bbox 3) border)) ; lower right
    expanded-bbox)) ;; Return the expanded bounding box

;; Purpose: Return the width of the bounding box.
(define (bbox-width bbox-in)
  (let ((bbox (ensure-vector bbox-in)))
    (- (vector-ref bbox 2) (vector-ref bbox 0))))

;; Purpose: Return the height of the bounding box.
(define (bbox-height bbox-in)
  (let ((bbox (ensure-vector bbox-in)))
    (- (vector-ref bbox 3) (vector-ref bbox 1))))

;; Purpose: Return the x offset of the bounding box.
(define (bbox-off-x bbox-in)
  (let ((bbox (ensure-vector bbox-in)))
    (vector-ref bbox 0)))

;; Purpose: Return the y offset of the bounding box.
(define (bbox-off-y bbox-in)
  (let ((bbox (ensure-vector bbox-in)))
    (vector-ref bbox 1)))

;; Purpose: Returns a bounding box clamped to fit within the image dimensions.
(define (crop-bbox-to-image-size bbox-in image)
  ;; bbox is a vector of the form #(left top right bottom)
  (let* ((bbox (ensure-vector bbox-in))
         (image-width  (image-get-width image))
         (image-height (image-get-height image))
         (left         (vector-ref bbox 0))
         (top          (vector-ref bbox 1))
         (right        (vector-ref bbox 2))
         (bottom       (vector-ref bbox 3))
         (clamped-left   (max 0 left))
         (clamped-top    (max 0 top))
         (clamped-right  (min image-width  right))
         (clamped-bottom (min image-height bottom)))
    (vector clamped-left clamped-top clamped-right clamped-bottom)))

;; Purpose: Returns a bounding box of the active image selection that is expanded
;;          by a border amount and clamped to the image dimensions.
(define (get-expanded-and-cropped-bbox image border)
  (let ((bbox (image-selection-get-bbox image))) ;; Get the initial bounding box
    (if bbox
      (crop-bbox-to-image-size (expand-bbox bbox border) image) ;; Expand and crop
      #f))) ;; Return #f if there is no selection

;; Purpose: Returns the layer id after copying an image all visible to a layer
(define (transfer-image-to-layer src-image dst-image target-layer mode)
  (gimp-selection-none src-image)
  (gimp-edit-copy-visible src-image)
  (paste-copied-to-drawable target-layer
                            dst-image
                            opaque
                            mode
                            TRUE))

;; Purpose: Returns the geometric mean of an image
(define (image-get-geometric-mean image)
  (sqrt (* (image-get-width image) (image-get-height image))))

;; =============================================================================
;; Purpose: Private Utilities
;;
;; This module provides a collection of private utility functions.
;; =============================================================================

;; =============================================================================
;; Purpose: GEGL helper functions > https://gegl.org/operations/
;;
;; This is a collection of functions designed to help apply GEGL filters
;;
;; =============================================================================

;; Purpose: Merge down a filter so that the filter is applied destructively
(define (filter-merge? layer filter merge)
  (if (boolean? merge) ; Ensure merge is a boolean
      (if merge
          (gimp-drawable-merge-filter layer filter)
          filter)
      (warning-message "Invalid merge value: expected #t or #f")))

;; Purpose: Creates a new GEGL filter
(define (gegl-filter layer gegl-name name)
  (let ((filter (car (gimp-drawable-filter-new layer gegl-name name))))
    filter))

;; Purpose: Converts opacity from the range [0, 100] to a normalized range [0, 1].
(define (gegl-opacity opacity)
  (* (validate-range opacity 0 100) 0.01))

;; =============================================================================
;; Filters

;; Purpose: Noise Spread
(define (gegl-noise-spread layer mode opacity amount-x amount-y seed merge verbose)
  (let ((filter (gegl-filter layer "gegl:noise-spread" "noise-spread"))
        (filter-config `((layer . ,(item-get-name layer))
                         (mode . ,mode)
                         (opacity . ,opacity)
                         (amount-x . ,amount-x)
                         (amount-y . ,amount-y)
                         (seed . ,seed)
                         (merge . ,merge))))
    (when verbose (debug-message "gegl-noise-spread: " filter-config))
    (gimp-drawable-filter-configure filter mode (gegl-opacity opacity)
                                    "amount-x" amount-x
                                    "amount-y" amount-y
                                    "seed" (msrg-rand))
    (filter-merge? layer filter merge)))

;; Purpose: Gaussian Blur
(define (gegl-gaussian-blur layer mode opacity std-dev-x std-dev-y filter-type merge verbose)
  (let ((filter (gegl-filter layer "gegl:gaussian-blur" "gaussian-blur"))
        (filter-config `((layer . ,(item-get-name layer))
                         (mode . ,mode)
                         (opacity . ,opacity)
                         (std-dev-x . ,std-dev-x)
                         (std-dev-y . ,std-dev-y)
                         (filter-type . ,filter-type)
                         (merge . ,merge))))
    (when verbose (debug-message "gegl:gaussian-blur: " filter-config))
    (gimp-drawable-filter-configure filter mode (gegl-opacity opacity)
                                    "std-dev-x" std-dev-x
                                    "std-dev-y" std-dev-y
                                    "filter" filter-type)
    (filter-merge? layer filter merge)))

